@@ -4,10 +4,9 @@ use WWW::Mechanize::PhantomJS;
 use Object::Import;
 use Carp qw(croak);
 
-use vars qw($VERSION @CARP_NOT);
-$VERSION= '0.22';
+our $VERSION= '0.23';
 
-@CARP_NOT = (qw[
+our @CARP_NOT = (qw[
     WWW::Mechanize::PhantomJS
 ]);
 
@@ -68,7 +67,7 @@ Max Maischein C<corion@cpan.org>
 
 =head1 COPYRIGHT (c)
 
-Copyright 2009-2015 by Max Maischein C<corion@cpan.org>.
+Copyright 2009-2020 by Max Maischein C<corion@cpan.org>.
 
 =head1 LICENSE
 
